# /packages/intranet-gtd-dashboard/tcl/intranet-gtd-dashboard-procs.tcl
#
# Copyright (c) 2003-2010 ]project-open[
#
# All rights reserved. Please check
# http://www.project-open.com/license/ for details.

ad_library {
    Library for ]po[ GTD dashboard
    @author klaus.hofeditz@project-open.com
}

namespace eval intranet-gtd-dashboard {}

ad_register_proc GET /intranet-gtd-dashboard/rest/* intranet-gtd-dashboard::dispatchRest
ad_register_proc POST /intranet-gtd-dashboard/rest/* intranet-gtd-dashboard::dispatchRest


ad_proc -public intranet-gtd-dashboard::dispatchRest {} {
    @return dispatches REST requests
    @author Klaus Hofeditz
} {
    set user_id [ad_maybe_redirect_for_registration]
    set urlpieces [ns_conn urlv]
    set path [lrange $urlpieces 2 [llength $urlpieces]]
    set url_query [ns_conn query]
    set header_vars [ns_conn form]
    set var_list [ad_ns_set_keys $header_vars]
    switch [lindex $urlpieces 2] {
        companies { return [intranet-gtd-dashboard::handle_rest_company [ns_conn method] $path $url_query $user_id ]  }
        timesheet { doc_return 200 "text/plain" [intranet-gtd-dashboard::handle_rest_timesheet [ns_conn method] $path $url_query $user_id ] }
        projects { doc_return 200 "text/plain" [intranet-gtd-dashboard::handle_rest_project [ns_conn method] $path $url_query $user_id] }
        action_items { doc_return 200 "text/plain" [intranet-gtd-dashboard::handle_action_items [ns_conn method] $path $url_query $user_id] }
        default { ad_return_complaint 1 "ressource not available"}
    }
}

ad_proc -public intranet-gtd-dashboard::handle_rest_timesheet {method path url_query user_id} {
    handles timesheet actions
} {

    # doc_return 200 "text/plain" [lindex $path 1]
        # doc_return 200 "text/plain" "$method $path $url_query $param"

    set query_list [split $url_query &]
    foreach sub_list $query_list {
	set query_item [split $sub_list = ]
	if {"scope" == [lindex $query_item 0] } {
	    set scope [lindex $query_item 1]
	} elseif { "hours" == [lindex $query_item 0] } {
	    set hours [lindex $query_item 1]
	} elseif { "hours_curr" == [lindex $query_item 0] } {
	    set hours_curr [lindex $query_item 1]
	}

    }
    set task_id [lindex $path 1]
    if { "GET"==$method } {
	if { "today"==$scope } {
	    doc_return 200 "text/plain" [db_string get_view_id "select hours from im_hours where day::date = now()::date and project_id=$task_id" -default 0]
	}
    } elseif { "POST"==$method } {
	if { "0"!=$hours_curr } {
               db_dml update "update im_hours set hours = $hours where day::date = now()::date and project_id=$task_id"
	} else {
	    set default_currency [ad_parameter -package_id [im_package_cost_id] "DefaultCurrency" "" "EUR"]
                set billing_rate 0
                set billing_currency ""

                db_0or1row get_billing_rate "
                        select  hourly_cost as billing_rate,
                        currency as billing_currency
                        from    im_employees
                        where   employee_id = $user_id
                "

	    if {"" == $billing_currency} { set billing_currency $default_currency }

	    set day_julian [db_string sysdate_as_julian "select to_char(sysdate,'J') from dual"]

                db_dml hours_insert "
                    insert into im_hours (
                        user_id, project_id,
                        day, hours,
                        billing_rate, billing_currency
                     ) values (
                        $user_id, $task_id,
                        to_date(:day_julian,'J'), $hours,
                        :billing_rate, :billing_currency
                     )"
	}
    }
}
ad_proc -public intranet-gtd-dashboard::handle_action_items {method path url_query user_id} {
    @return project list / task list
} {

    set header_vars [ns_conn form]
    set var_list [ad_ns_set_keys $header_vars]
    ad_ns_set_to_tcl_vars $header_vars

    if { "" != $item_list } {
        return [intranet-gtd-dashboard::write_gtd_list $item_list $user_id]
    } else {
        return "error: not items found"
    }
}

ad_proc -public intranet-gtd-dashboard::render_json { object_list search_string } {
    @tkes a list of objects returns them as a json structure
} {
    #

    set output "\{\"ResultSet\":\n\{\n\"Result\": \[\n"
    #     set output "\[\n"

    foreach sub_list $object_list {
        if { [llength $sub_list] } {
            if { 0 != [llength [lindex $sub_list 0]] } {
                if { "" != $search_string } {
		    if { [string first  [string tolower $search_string] [string tolower [lindex $sub_list 0]]] != -1 } {
			append output "\{\"Object\":\""
			append output [string map {&nbsp; ""}  [lindex $sub_list 0] ]
                            append output "\""
                            append output ",\"ObjectId\":"
			append output [string map {&nbsp; ""}  [lindex $sub_list 1] ]
			#                           append output [lindex $sub_list 1]
			append output "\},"
		    }
                } else {
                    append output "\{\"Object\":\""
                    append output [string map {&nbsp; ""}  [lindex $sub_list 0] ]
                    append output "\"\},"
                }
            }
        }
    }

    set output "[string range $output 0 [expr [string length $output]-2]]"
    append output "\]\}\n\}"
    #    append output "\n\]"
}


ad_proc -public intranet-gtd-dashboard::handle_rest_project {method path url_query user_id} {
    @return project list / task list
} {


    set query_list [split $url_query &]


    # find searchstring, object type, project_id
    foreach sub_list $query_list {
        set query_item [split $sub_list = ]
        if {"search_string" == [lindex $query_item 0] } {
	    set search_string [lindex $query_item 1]
        }
        # project or task?
        if {"object_type" == [lindex $query_item 0] } {
            set object_type [lindex $query_item 1]
        }
        if {"project_id" == [lindex $query_item 0] } {
            set project_id [lindex $query_item 1]
        }
        # last id used in <li> element
        if {"last_id" == [lindex $query_item 0] } {
            set last_id [lindex $query_item 1]
        }
        # location
        if {"location" == [lindex $query_item 0] } {
            set location [lindex $query_item 1]
        }

    }

    # set project_id 27971

    # Getting list of objects

    switch $object_type {
        project {
            set object_list [im_project_list -exclude_subprojects_p 0 -exclude_status_id [im_project_status_closed] -project_id 0]
            set output [intranet-gtd-dashboard::render_json $object_list $search_string]
        }
        task {
            set output [intranet-gtd-dashboard::render_output [im_gtd_task_list -restrict_to_project_id $project_id] task_table $last_id ts $location]
        }
        default {set output "Object Type not found"}
    }
    return $output

}


# ----------------------------------------------------------------------
# PM helper functions
# ----------------------------------------------------------------------


ad_proc -public intranet-gtd-dashboard::read_gtd_list { user_id} {
    @reads gtd list
} {


set sql "
	SELECT 
	      p.project_id as task_id,
	      p.project_name as task_name
	FROM 
		im_gtd_tasks t 
	LEFT JOIN 
		im_projects p
	ON 
		t.ref_id=p.project_id
	WHERE
		t.owner_id = $user_id
	ORDER BY
	      t.gtd_task_id
"
    set options [list]
    db_foreach task $sql {
        lappend options [list $task_name $task_id]
    }
    return $options
}

ad_proc -public intranet-gtd-dashboard::write_gtd_list { object_list user_id} {
    @writes gtd list 
} {
	# delete current items
	db_dml delete "
                delete from im_gtd_tasks;
	"
	set object_list [split $object_list ,]

	foreach sub_list $object_list {
		set item [split $sub_list _]
		set item_ref [lindex $item 2]
		set item_type [lindex $item 3]
		if { "ts" == $item_type } {
			set item_type 41200
		} else {
			set item_type 0
		}
		db_dml insert_gtd "insert into im_gtd_tasks (gtd_task_id,ref_id,owner_id,created_date,gtd_type_id,gtd_task_status_id) values (nextval('im_gtd_tasks_seq'),$item_ref,$user_id,now(),$item_type,41300)"
	}
}


ad_proc -public intranet-gtd-dashboard::render_output { object_list template last_id type location} {
    @returns task html snippet created based on JSON object list 
} {
    set ctr [expr $last_id*1]
	set output "<ul id=\"$location\" class=\"draglist\">"
	foreach sub_list $object_list {
	    if { [llength $sub_list] } {
		append output " <li class=\"list0\" id=\"li999_$ctr\">"
		append output "         <table class=\"table_task_item\" name=\""
		append output "li999_$ctr"
		append output "_"
		append output [lindex $sub_list 1]
		append output "_$type"
		append output "\">"
		append output "                 <tr>"
		append output "                         <td width='40px'>[lindex $sub_list 1]</td>"
		append output "                         <td>[lindex $sub_list 0]</td>"
		# append output "                         <td width=\"15%\"></td>"
		# append output "                         <td width=\"15%\"></td>"
		append output "                         <td width='20px' align='right'><button id=\"show\" name=\""
		append output [lindex $sub_list 1]
		append output "\"><img alt=\"\" src=\"/intranet/images/clock.gif\" height='16' width='16' /></button></td>"
		append output "                 </tr>"
		append output "         </table>"
		append output " </li>"
	    }
	    incr ctr 
	}
    append output "</ul>"
    return $output
}




ad_proc -public im_project_list { 
    {-include_project_ids {} }
    {-exclude_subprojects_p 1}
    {-exclude_tasks_p 1}
    {-exclude_status_id ""}
    {-exclude_type_id ""}
    {-project_status_id 0}
    {-project_type_id 0}
    {-member_user_id 0}
    {-company_id 0}
    {-project_id 0}
    {-query_string ""}
} { 
    Get a list of projects, only projects 
    Filter is based on on_track_status_id, which is an attribute only projects do have 
} {

    # Default: Exclude tasks and deleted projects
    if {"" == $exclude_status_id} { set exclude_status_id [im_project_status_deleted] }
    if {!$exclude_tasks_p} { set exclude_subprojects_p 0 }

    set current_project_id $project_id
    set current_user_id [ad_get_user_id]
    set max_project_name_len 50

    # Make sure we don't get a syntax error in the query
    lappend include_project_ids 0

    set list_sort_order [parameter::get_from_package_key -package_key "intranet-timesheet2" -parameter TimesheetAddHoursSortOrder -default "name"]

    # Handle subprojects 

    set subsubproject_sql ""
    set subprojects [list 0]
    if {0 != $current_project_id} {

	# Determine the topmost project in the hierarchy
	set super_project_id $current_project_id
	set loop 1
	set ctr 0
	while {$loop && $ctr < 100} {
	    set loop 0
	    set parent_id [db_string parent_id "
		select parent_id 
		from im_projects
		where project_id = :super_project_id
	    " -default ""]
	    if {"" != $parent_id} {
		set super_project_id $parent_id
		set loop 1
	    }
	    incr ctr
	}

	# Check permissions for showing subprojects
	set perm_sql "
		(select p.*
		from    im_projects p,
			acs_rels r
		where   r.object_id_one = p.project_id
			and r.object_id_two = :current_user_id
		)
	"
	if {[im_permission $current_user_id "view_projects_all"]} { set perm_sql "im_projects" }


	set subprojects [db_list subprojects "
		select	children.project_id
		from	im_projects parent,
			$perm_sql children
		where
			children.tree_sortkey 
				between parent.tree_sortkey 
				and tree_right(parent.tree_sortkey)
			and children.project_type_id not in (
				84, [im_project_type_task]
			)
			and parent.project_id = :super_project_id

			-- exclude the projects own subprojects
			-- to avoid circular loops
			and children.project_id not in (
				select	subchild.project_id
				from	im_projects subparent,
					im_projects subchild
				where
					subchild.tree_sortkey 
						between subparent.tree_sortkey 
						and tree_right(subparent.tree_sortkey)
					and subparent.project_id = :current_project_id
			)
	"]

	# Add an invalid project in order to avoid an empty list of subprojects
	# and a resulting SQL syntax error in "parent_id in ()"
	lappend subprojects 0

    }

    # ---------------------------------------------------------
    # Compile "criteria"

    set p_criteria [list]
    set main_p_criteria [list]
    if {$exclude_subprojects_p} { 
	lappend p_criteria "p.parent_id is null" 
    }
   
    if {0 != $company_id && "" != $company_id} { 
	lappend p_criteria "p.company_id = :company_id" 
	# Main project should have same customer, but don't enforce it!
    }

    if {0 != $exclude_status_id && "" != $exclude_status_id} {
	lappend p_criteria "p.project_status_id not in ([join [im_sub_categories $exclude_status_id] ","])"
	lappend main_p_criteria "p.project_status_id not in ([join [im_sub_categories $exclude_status_id] ","])"
    }

    if {0 != $exclude_type_id && "" != $exclude_type_id} {
	lappend p_criteria "p.project_type_id not in ([join [im_sub_categories $exclude_type_id] ","])"
	# No restriction of type on parent project!
    }

    if {$exclude_tasks_p} {
	lappend p_criteria "p.project_type_id not in ([join [im_sub_categories [im_project_type_task]] ","])"
	# Main project is never of type task...
    }

    if {0 != $project_status_id && "" != $project_status_id} {
	lappend p_criteria "p.project_status_id in ([join [im_sub_categories $project_status_id] ","])"
	# No restriction on parent's status id
    }

    if {0 != $project_type_id && "" != $project_type_id} {
	lappend p_criteria "p.project_type_id in ([join [im_sub_categories $project_type_id] ","])"
	# No restriction on parent's project type!
    }

    if {0 != $member_user_id && "" != $member_user_id} {
	lappend p_criteria "p.project_id in (
					select	object_id_one
					from	acs_rels
					where	object_id_two = :member_user_id
	)"
	# No restriction on parent project membership, because parent
	# projects always have the same members as sub-projects.
    }

    # Unprivileged members can only see the projects they're participating
    if {![im_permission $current_user_id view_projects_all]} {
	lappend p_criteria "p.project_id in (
					select	object_id_one
					from	acs_rels
					where	object_id_two = :current_user_id
	)"
	# No restriction on parent project membership, because parent
	# projects always have the same members as sub-projects.
    }


    # -----------------------------------------------------------------
    # Compose the SQL

    set p_where_clause [join $p_criteria " and\n\t\t\t\t\t"]
    if { ![empty_string_p $p_where_clause] } {
	set p_where_clause " and $p_where_clause"
    }

    set main_p_where_clause [join $p_criteria " and\n\t\t\t\t\t"]
    if { ![empty_string_p $main_p_where_clause] } {
	set main_p_where_clause " and $main_p_where_clause"
    }

    switch $list_sort_order {
	name { set sort_order "lower(p.project_name)" }
	order { set sort_order "p.sort_order" }
	legacy { set sort_order "p.tree_sortkey" }
	default { set sort_order "lower(p.project_nr)" }
    }

    set sql "
		select
			p.project_id,
			p.parent_id,
			tree_level(p.tree_sortkey) as tree_level,
			substring(p.project_name for :max_project_name_len) as project_name_shortened,
			$sort_order as sort_order,
			p.on_track_status_id
		from
			im_projects p,
			im_projects main_p,
			(	select	p.project_name,
					p.project_id
				from	im_projects p
				where	1=1
					$p_where_clause
			    UNION
				select	p.project_name,
					p.project_id
				from	im_projects p
				where	p.project_id in ([join $subprojects ", "])
			    UNION
				select	p.project_name,
					p.project_id
				from	im_projects p
				where	p.project_id = :current_project_id
			    UNION
				select	p.project_name,
					p.project_id
				from	im_projects p
				where	p.project_id in ([join $include_project_ids ","])
			) p_cond,
			(	select	p.project_id
				from	im_projects p
				where	1=1
					$p_where_clause
			) main_p_cond
		where
			p.on_track_status_id is NOT null and
			p.project_id = p_cond.project_id and
			main_p.project_id = main_p_cond.project_id and
			main_p.parent_id is null and
			tree_ancestor_key(p.tree_sortkey, 1) = main_p.tree_sortkey and
			main_p.project_status_id not in ([im_project_status_deleted]) and
			p.project_status_id not in ([im_project_status_deleted]) and
 			p.project_name like '%$query_string%'
		order by 
			lower(main_p.project_name),
			p.tree_sortkey
    "

    set options [list]
    db_foreach project_name $sql {
	set indent ""
	for {set i 0} {$i < $tree_level} { incr i} { append indent "&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;" }
	lappend options [list "$indent$project_name_shortened" $project_id]
    }
    return $options
}





ad_proc -public im_gtd_task_list {
    {-order_by "priority"} 
    {-restrict_to_type_id 0} 
    {-restrict_to_status_id 0} 
    {-restrict_to_material_id 0} 
    {-restrict_to_project_id 0} 
    {-restrict_to_mine_p "all"} 
    -current_page_url 
    -return_url 
} {
    returns a list of tasks 
} {
    # ---------------------- Security - Show the comp? -------------------------------
    set user_id [ad_get_user_id]

    # Is this a "Consulting Project"?
#    if {0 != $restrict_to_project_id} {
#	if {![im_project_has_type $restrict_to_project_id "Consulting Project"]} {
#	    return $restrict_to_project_id
#	}
#    }
    
    # Check vertical permissions - Is this user allowed to see TS stuff at all?
    if {![im_permission $user_id "view_timesheet_tasks"]} {
	return ""
    }

    # Check if the user can see all timesheet tasks
    if {![im_permission $user_id "view_timesheet_tasks_all"]} {
	set restrict_to_mine_p "mine"
    }


    # Check horizontal permissions -
    # Is the user allowed to see this project?
    im_project_permissions $user_id $restrict_to_project_id view read write admin
    if {!$read && ![im_permission $user_id view_timesheet_tasks_all]} { return ""}

    # Check for Timesheet tasks of a certain status.
    # The status_id is only available in the previous screen.
    # Very Ugly!!

    # fraber 090318 @ phast: Causes problems, need to disable
#    upvar subproject_status_id subproject_status_id
#    if {![info exists subproject_status_id]} { set subproject_status_id 0 }
#    if {"" == $subproject_status_id} { set subproject_status_id 0 }
#    set subproject_sql ""
#    if {$subproject_status_id} {
#	set subproject_sql "and p.project_status_id in ([join [im_sub_categories $subproject_status_id] ","])"
#    }
    set subproject_sql ""

    # -------- Get parameters from HTTP session -------
    # Don't trust the container page to pass-on that value...

    set form_vars [ns_conn form]
    if {"" == $form_vars} { set form_vars [ns_set create] }

    set start_idx [ns_set get $form_vars "task_start_idx"]
    if {"" == $start_idx} { set start_idx 0 }

    # ---------------------- Defaults ----------------------------------

    set date_format "YYYY-MM-DD"


    # ---------------------- Build the SQL query ---------------------------
    set order_by_clause "order by p.project_nr, t.task_id"
    set order_by_clause_ext "order by project_nr, task_name"

    switch $order_by {
	"Status" { 
	    set order_by_clause "order by t.task_status_id" 
	    set order_by_clause_ext "m.task_id"
	}
    }
	
    set criteria [list]

    set project_restriction "t.project_id = :restrict_to_project_id"

    lappend criteria $project_restriction

    if {[string is integer $restrict_to_status_id] && $restrict_to_status_id > 0} {
	lappend criteria "t.task_status_id in ([join [im_sub_categories $restrict_to_status_id] ","])"
    }

    if {"mine" == $restrict_to_mine_p} {
	lappend criteria "t.task_id in (select object_id_one from acs_rels where object_id_two = [ad_get_user_id])"
    }

    if {[string is integer $restrict_to_type_id] && $restrict_to_type_id > 0} {
	lappend criteria "t.task_type_id in ([join [im_sub_categories $restrict_to_type_id] ","])"
    }

    set restriction_clause [join $criteria "\n\tand "]
    if {"" != $restriction_clause} { 
	set restriction_clause "and $restriction_clause" 
    }


    # ---------------------- Inner Permission Query -------------------------

    # Check permissions for showing subprojects
    set children_perm_sql "
        (select p.*
        from    im_projects p,
                acs_rels r
        where   r.object_id_one = p.project_id
                and r.object_id_two = :user_id
        )
    "

    if {[im_permission $user_id "view_projects_all"]} { 
	set children_perm_sql "
	(select	t.*
	 from	im_projects t
	 where	$project_restriction
	)
	"
    }

    set projects_perm_sql "
	(select	t.*
	from	im_projects t,
		acs_rels r
	where	r.object_id_one = t.project_id
		and r.object_id_two = :user_id
		and $project_restriction
	)
    "

    if {[im_permission $user_id "view_projects_all"]} {
	set projects_perm_sql "
	(select	t.*
	 from	im_projects t
	 where	$project_restriction
	)
	"
    }

    # ---------------------- Get the SQL Query -------------------------

set sql "	
    select
    	t.*,
	    p.project_id,
	    p.project_name,
	    p.project_nr,
	    p.note,
	    cc.cost_center_name,
	    cc.cost_center_code,
	    im_category_from_id(t.task_type_id) as task_type,
	    im_category_from_id(t.task_status_id) as task_status,
	    im_category_from_id(t.uom_id) as uom,
	    im_material_nr_from_id(t.material_id) as material_nr,
	    to_char(t.percent_completed, '999990') as percent_completed_rounded
    from
	    $projects_perm_sql p,
            im_timesheet_tasks_view t
	    left outer join im_cost_centers cc on (t.cost_center_id = cc.cost_center_id)
    where
	    t.project_id = p.project_id
	    $subproject_sql
	    $restriction_clause
	    $order_by_clause
"
    set options [list]
    db_foreach task $sql {
	lappend options [list $task_name $task_id]
    }
    return $options
}




# if {"" == $package_gtd_dashboard_id {} {
#    Returns the package id of the intranet-gtd-dashboard module
# } {
#    return [util_memoize "im_package_intranet_gtd_dashboard_id_helper"]
# }




ad_proc -private im_package_intranet_gtd_dashboard_id_helper {} {
    return [db_string im_package_core_id {
        select package_id from apm_packages
        where package_key = 'intranet-gtd-dashboard'
    } -default 0]
}




